
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "shared.h"
#include "semaphore.h"


#define BLOCK_SIZE 64
#define BUFFER_SIZE 16

#define FALSE 0
#define TRUE 1

// struttura di definizione dell'area di memoria condivisa
struct SHARED
{
 unsigned char buffer[BLOCK_SIZE*BUFFER_SIZE]; // buffer blocchi dati
 unsigned int count; // numero di blocchi dati nel buffer
 int end; // flag di segnalazione della fine dei dati
};

FILE *input_file, *output_file; // identificatori file input/output
int space_semaphore; // semaforo per segnalazione spazio nel buffer
int data_semaphore; // semaforo per segnalazione dati nel buffer
int mutex_semaphore; // semaforo per accesso mutuamente esclusivo
struct SHARED *shared; // puntatore area di memoria condivisa
int buffer_write_index; // indice per scrittura blocchi nel buffer
int buffer_read_index; // indice per lettura blocchi dal buffer


void producer(void)
{
 int n;
 char buffer[BLOCK_SIZE];

 do
   {
    semaphore_wait(space_semaphore);
    // lettura dati da file di input e copia nel buffer condiviso
    n = fread(buffer, 1, BLOCK_SIZE, input_file);
    memset( &(shared->buffer[buffer_write_index*BLOCK_SIZE]), 0, BLOCK_SIZE);
    memcpy( &(shared->buffer[buffer_write_index*BLOCK_SIZE]), buffer, n);
    // aggiornamento indice di scrittura nel file
    buffer_write_index = (buffer_write_index+1) % BUFFER_SIZE;
    // aggiornamento numero blocchi dati e flag fine dati
    semaphore_wait(mutex_semaphore);
    shared->count++;
    if (feof(input_file))
      shared->end = TRUE;
    semaphore_signal(mutex_semaphore);
    semaphore_signal(data_semaphore);
   } while (!feof(input_file));
 // segnalazione per consumatore in attesa
 semaphore_signal(data_semaphore);
}


void consumer(void)
{
 int count, end;

 do
   {
    semaphore_wait(data_semaphore);
    // verifica effettiva disponibilita' dati
    semaphore_wait(mutex_semaphore);
    count = shared->count;
    end = shared->end;
    semaphore_signal(mutex_semaphore);
    if (end && count==0)
      break;
    // scrittura blocco-dati del buffer condiviso nel file di output
    fwrite( &(shared->buffer[buffer_read_index*BLOCK_SIZE]), 1, BLOCK_SIZE, output_file);
    // aggiornamento indice di scrittura nel file
    buffer_read_index = (buffer_read_index+1) % BUFFER_SIZE;
    // aggiornamento numero blocchi dati e copia flag fine dati
    semaphore_wait(mutex_semaphore);
    count = shared->count--;
    end = shared->end;
    semaphore_signal(mutex_semaphore);
    semaphore_signal(space_semaphore);
   } while (!end || count>0);
}




int main(int argc, char* argv[])
{
 int pid;
 
 if (argc != 3)
   {
    printf("Uso: %s input-file output-file\r\n",argv[0]);
    return 0;
   }
 // allocazione area memoria condivisa
 shared = (struct SHARED*)shared_memory_alloc(sizeof(struct SHARED));
 if (shared == (struct SHARED*)(-1))
   {
    printf("shared_memory_alloc error\r\n");
    return 0;
   }
 // inizializzazione semafori
 space_semaphore = semaphore_init(BUFFER_SIZE);
 if (space_semaphore == -1)
   {
    printf("semaphore_init error\r\n");
    shared_memory_free(shared);
    return 0;
   }
 data_semaphore = semaphore_init(0);
 if (data_semaphore == -1)
   {
    printf("semaphore_init error\r\n");
    semaphore_remove(space_semaphore);
    shared_memory_free(shared);
    return 0;
   }
 mutex_semaphore = semaphore_init(1);
 if (mutex_semaphore == -1)
   {
    printf("semaphore_init error\r\n");
    semaphore_remove(space_semaphore);
    semaphore_remove(data_semaphore);
    shared_memory_free(shared);
    return 0;
   }
 // apertura file input/output in lettura/scrittura
 input_file = fopen(argv[1], "rb");
 if (input_file == NULL)
   {
    printf("input file open error\r\n");
    semaphore_remove(space_semaphore);
    semaphore_remove(data_semaphore);
    semaphore_remove(mutex_semaphore);
    shared_memory_free(shared);
    return 0; 
   }
 output_file = fopen(argv[2], "wb");
 if (output_file == NULL)
   {
    printf("output file open error\r\n");
    semaphore_remove(space_semaphore);
    semaphore_remove(data_semaphore);
    semaphore_remove(mutex_semaphore);
    shared_memory_free(shared);
    fclose(input_file);
    return 0;
   }
 // inizializzazione variabili condivise e indici del buffer
 shared->count = 0;
 shared->end = FALSE;
 buffer_write_index = 0;
 buffer_read_index = 0;
 // creazione processi produttore e consumatore
 pid = fork();
 if (pid < 0)
   {
    printf("fork error\r\n");
    semaphore_remove(space_semaphore);
    semaphore_remove(data_semaphore);
    semaphore_remove(mutex_semaphore);
    shared_memory_free(shared);
    fclose(input_file);
    fclose(output_file);
    return 0;
   }
 if (pid > 0)
   { 
    // processo padre (produttore)
    producer();
    wait(NULL); // attesa terminazione processo figlio
   }
 else
     {
      // processo figlio (consumatore)
      consumer();
      exit(1); // terminazione processo figlio
     }
 shared_memory_free(shared);
 semaphore_remove(space_semaphore);
 semaphore_remove(data_semaphore);
 semaphore_remove(mutex_semaphore);
 fclose(input_file);
 fclose(output_file);
 return 1;
}
